"""
The PDA ("main" class) for the :mod:`royalnet_telethon` frontend.
"""

from __future__ import annotations

import enum
import logging

import discord
import discord as d
import royalnet.engineer as engi
import royalnet.royaltyping as t

from .bullet.projectiles import DiscordMessageReceived, DiscordMessageEdited, DiscordMessageDeleted

log = logging.getLogger(__name__)


class DiscordpyPDAMode(enum.Enum):
    """
    .. todo:: Document this.
    """

    GLOBAL = enum.auto()
    CHANNEL = enum.auto()
    USER = enum.auto()
    CHANNEL_USER = enum.auto()


class DiscordpyPDAImplementation(engi.ConversationListImplementation):
    """
    .. todo:: Document this.
    """

    @property
    def namespace(self):
        return "discordpy"

    def __init__(self, name: str, bot_token: str,
                 mode: DiscordpyPDAMode = DiscordpyPDAMode.CHANNEL_USER,
                 intents: discord.Intents = discord.Intents.default()):

        super().__init__(name=name)

        self.mode: DiscordpyPDAMode = mode
        """
        The mode to use for mapping dispensers.
        """

        self.bot_token: str = bot_token
        """
        .. todo:: Document this.
        """

        # noinspection PyMethodParameters
        class CustomClient(d.Client):
            async def on_ready(cli):
                log.debug("CustomClient is ready!")

            async def on_error(self, event_method, *args, **kwargs):
                log.error(f"An error occoured in CustomClient: {event_method!r} {args!r} {kwargs!r}")

            async def on_message(cli, message: d.Message):
                log.debug("Triggered on_message, putting in dispenser...")

                await self.put(
                    key=self._determine_key(message=message),
                    projectile=DiscordMessageReceived(event=message)
                )

            async def on_message_edit(cli, message: d.Message):
                log.debug("Triggered on_message_edit, putting in dispenser...")

                await self.put(
                    key=self._determine_key(message=message),
                    projectile=DiscordMessageEdited(event=message)
                )

            async def on_message_delete(cli, message: d.Message):
                log.debug("Triggered on_message_delete, putting in dispenser...")

                await self.put(
                    key=self._determine_key(message=message),
                    projectile=DiscordMessageDeleted(event=message)
                )

        self.client: d.Client = CustomClient(intents=intents)
        """
        .. todo:: Document this.        
        """

    def _determine_key(self, message: d.Message):
        """
        .. todo:: Document this.
        """

        if self.mode == DiscordpyPDAMode.GLOBAL:
            return None
        elif self.mode == DiscordpyPDAMode.USER:
            author: d.User = message.author
            return author.id
        elif self.mode == DiscordpyPDAMode.CHANNEL:
            channel: t.Union[d.DMChannel, d.TextChannel] = message.channel
            return channel.id
        elif self.mode == DiscordpyPDAMode.CHANNEL_USER:
            author: d.User = message.author
            channel: t.Union[d.DMChannel, d.TextChannel] = message.channel
            return author.id, channel.id
        else:
            raise TypeError("Invalid mode")

    async def run(self) -> t.NoReturn:
        await self.client.start(self.bot_token, bot=True, reconnect=True)


__all__ = (
    "DiscordpyPDAMode",
    "DiscordpyPDAImplementation",
)



from __future__ import annotations

import datetime

import async_property as ap
import discord
import discord as d
import royalnet.engineer.bullet.contents as co
import royalnet.royaltyping as t

from royalnet_discordpy.formatting import ds_markdown_format


class DiscordMessage(co.Message):
    def __init__(self, msg: d.Message):
        super().__init__()
        self._msg: d.Message = msg

    def __hash__(self) -> int:
        return self._msg.id

    @ap.async_property
    async def text(self) -> t.Optional[str]:
        return self._msg.content

    @ap.async_property
    async def timestamp(self) -> t.Optional[datetime.datetime]:
        return max(self._msg.date, self._msg.edit_date)

    @ap.async_property
    async def reply_to(self) -> t.Optional[DiscordMessage]:
        return DiscordMessage(msg=self._msg.reference.cached_message)

    @ap.async_property
    async def channel(self) -> t.Optional[DiscordChannel]:
        return DiscordChannel(channel=self._msg.channel)

    @ap.async_property
    async def sender(self) -> t.Optional[DiscordUser]:
        sender: t.Union[d.User, d.Member] = self._msg.author
        return DiscordUser(user=sender)

    async def reply(self, *,
                    text: str = None,
                    files: t.List[t.BinaryIO] = None) -> t.Optional[DiscordMessage]:
        if files is None:
            files = []

        msg = await self._msg.reply(content=ds_markdown_format(text) if text else None,
                                    files=[discord.File(file) for file in files])
        return DiscordMessage(msg=msg)


class DiscordChannel(co.Channel):
    def __init__(self, channel: t.Union[d.DMChannel, d.TextChannel, d.GroupChannel]):
        super().__init__()
        self._channel: t.Union[d.DMChannel, d.TextChannel, d.GroupChannel] = channel

    def __hash__(self):
        return self._channel.id

    @ap.async_property
    async def name(self) -> t.Optional[str]:
        return self._channel.name

    @ap.async_property
    async def topic(self) -> t.Optional[str]:
        return self._channel.topic

    @ap.async_property
    async def users(self) -> t.List[DiscordUser]:
        return [DiscordUser(user=member) for member in self._channel.members]

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> t.Optional[DiscordMessage]:
        if files is None:
            files = []

        msg = await self._channel.send(content=ds_markdown_format(text), files=[discord.File(file) for file in files])
        return DiscordMessage(msg=msg)


class DiscordUser(co.User):
    def __init__(self, user: t.Union[d.User, d.Member]):
        super().__init__()
        self._user: t.Union[d.User, d.Member] = user

    def __hash__(self):
        return self._user.id

    @ap.async_property
    async def name(self) -> t.Optional[str]:
        return f"{self._user.mention}"

    async def slide(self) -> DiscordChannel:
        dm = self._user.dm_channel
        if dm is None:
            dm = await self._user.create_dm()
        return DiscordChannel(channel=dm)


__all__ = (
    "DiscordMessage",
    "DiscordChannel",
    "DiscordUser",
)



def ds_markdown_format(string: str) -> str:
    """
    Format a string as Discord Markdown.

    The following characters will be considered control characters:
    - ``\uE01B``: start bold
    - ``\uE00B``: end bold
    - ``\uE011``: start italic
    - ``\uE001``: end italic
    - ``\uE012``: start underline
    - ``\uE002``: end underline
    - ``\uE015``: start strike
    - ``\uE005``: end strike
    - ``\uE01F``: start spoiler
    - ``\uE00F``: end spoiler
    - ``\uE01C``: start single-line code
    - ``\uE00C``: end single-line code
    - ``\uE01D``: start multi-line code
    - ``\uE00D``: end multi-line code

    :param string: The string to format.
    :return: The formatted string.

    .. warning:: For now, this is a Discordpy implementation detail.

    .. todo:: This may cause denial of service attacks from users!
    """

    string = string.replace("*", "\\*")
    string = string.replace("_", "\\_")
    string = string.replace("`", "\\`")
    string = string.replace("~", "\\~")

    string = string.replace("\uE01B", "**")
    string = string.replace("\uE00B", "**")
    string = string.replace("\uE011", "_")
    string = string.replace("\uE001", "_")
    string = string.replace("\uE015", "~~")
    string = string.replace("\uE005", "~~")
    string = string.replace("\uE012", "__")
    string = string.replace("\uE002", "__")
    string = string.replace("\uE01F", "||")
    string = string.replace("\uE00F", "||")
    string = string.replace("\uE01C", "`")
    string = string.replace("\uE00C", "`")
    string = string.replace("\uE01D", "```")
    string = string.replace("\uE00D", "```")
    return string
